/*

	Mirroring audio driver
	Audio pipe
	by P. Starzetz '2002

	Idea: what you write to the one end will appear on another end as sound data
		and vice versa


	How:
	1) we allocate two audio devices (dsp) using some buffer space
	2) the read buffer of the first device is the output buffer of the second device
	3) the data i/o is driven by a timer
	4) it should be possible to always read & write to the device even if no readers are there
	5) the timer routine creates/discards audio data at the configured sample rate

	Todo:
	- it is full of race conditions and zasadzkas:-))
	- support more ioctls()
	- support dynamic device allocation/deallocation
	- support different sample rates & resampling
	- support a mixer device
	- reorganize parts of the code to remove the gotoart
	- support /proc
	- make it SMP safe

	Pitfalls:
	- do not kmalloc from timer interrupt!

	Revision history:
	- ....
	- 0.18 first stable version of two dsp driver
	- 0.19 first kernel thread driven version of the driver
	- 0.20 first stable kernel thread driven version

*/




#include <linux/version.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/sound.h>
#include <linux/slab.h>
#include <linux/soundcard.h>
#include <linux/pci.h>
#include <linux/bitops.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/spinlock.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>
#include <asm/atomic.h>
#include <asm/processor.h>



#define MODMSG KERN_INFO MNAME " " MVERSION ": "
#define MODERR KERN_ERR MNAME " " MVERSION ": "
#define MADEBUG if(madebug)
#define EXPORT_ALL_SYMBOLS


#define MA_MAGIC_NUM		0x91827498


#define MA_HARD_MAXFRAGSIZE 	4096
#define MA_HARD_MINFRAGSIZE 	1024
#define MA_HARD_MAXFRAGS	16
#define MA_HARD_MINRATE		8000
#define MA_HARD_MAXRATE		44100


#define MA_TIMER_STOP		(0x01u << 0)
#define MA_TIMER_KILL_THREAD	(0x01u << 1)

#define MA_TIMER_STATICS	6000





static int madebug = 0;

static int printstat = 0;




struct maudio_state;
typedef struct maudio_state MState;



wait_queue_head_t dsp_open_wait;
wait_queue_head_t psd_open_wait;

DECLARE_WAIT_QUEUE_HEAD (maudiod_wait);


static MState dsp_state;
static MState psd_state;


struct timer_list matim;

static volatile int maticks;

static volatile unsigned matimerflg;

static volatile pid_t kthrpid;



struct MABuff
{

  struct MABuff *next;
  struct MABuff *prev;

  union
  {
    unsigned char *cbf;
    unsigned short *sbf;
    unsigned int *wbf;
  }
  buf;

  unsigned buffersize;

  unsigned bytes;

  unsigned coff;

  int mbcpos;

};

typedef struct MABuff MABuff;



typedef struct
{

  MABuff *head;			//      fake head
  MABuff *tail;			//      fake tail


  volatile unsigned mbc_size;	// how many cache entries, current fragment limit, must be less than maxfrags, can be set with SETFRAGMENTS
  MABuff **mbcache;

  struct semaphore mbcsem;

  int fragsize;

  unsigned maxfrags;
  unsigned maxfragsize;
  unsigned minfragsize;

  volatile int frags;

  volatile int totalbytes;

  size_t count;

  volatile unsigned ticks;

  volatile int rtok;

  int wakeup_thr;

  wait_queue_head_t read_wait, write_wait;

  volatile MState *read_dev;

  volatile MState *write_dev;

  struct semaphore devsem;


}
MBFifo;


struct maudio_state
{

  unsigned int magic;

  MBFifo *rmf;

  MBFifo *wmf;

  volatile int opens;

  int dev_audio;
  int dev_mixer;

  unsigned fmt;

  unsigned channels;

  unsigned samplerate;

  unsigned depth;

  volatile unsigned bytespersec;

  mode_t open_mode;
  wait_queue_head_t *open_wait;
  struct semaphore opensem;
};


static MBFifo global_readf, global_writef;


MState *
ms_init_device (MState * ms)
{
  ms->fmt = AFMT_U16_LE;
  ms->fmt = 0x10;
  ms->channels = 2;
  ms->samplerate = 44100;
  ms->depth = 2;
  ms->bytespersec = ms->channels * ms->depth * ms->samplerate;

  ms->rmf = NULL;
  ms->wmf = NULL;

  ms->magic = MA_MAGIC_NUM;

  return ms;
}


MState *
ms_cleanup_device (MState * ms)
{

  return ms;
}


MABuff *
mf_insert_mabuff (MBFifo * mf, int pos)
{
  MABuff *ma = NULL;


  if (pos < 0 || pos > mf->mbc_size)
    {
      return NULL;
    }

  ma = (MABuff *) kmalloc (sizeof (MABuff), GFP_KERNEL);

  if (!ma)
    {
      return NULL;
    }

  ma->buf.cbf = (char *) kmalloc (mf->fragsize, GFP_KERNEL);
  if (!ma->buf.cbf)
    {
      kfree (ma);
      return NULL;
    }

  ma->buffersize = mf->fragsize;
  ma->bytes = 0;
  ma->coff = 0;
  ma->prev = NULL;
  ma->next = NULL;

  ma->mbcpos = pos;
  mf->mbcache[pos] = ma;

  return ma;
}


MBFifo *
mf_cache_cleanup (MBFifo * mf)
{
  int i = 0;


  if (!mf->mbcache)
    {
      return NULL;
    }


  for (i = 0; i < mf->mbc_size; i++)
    {
      if (mf->mbcache[i] != NULL)
	{
	  if (mf->mbcache[i]->buf.cbf)
	    {
	      kfree ((mf->mbcache[i]->buf).cbf);
	      (mf->mbcache[i]->buf).cbf = NULL;
	    }
	  else
	    kfree (mf->mbcache[i]);
	  mf->mbcache[i] = NULL;
	}
    }

  kfree (mf->mbcache);

  mf->head->next = mf->tail;
  mf->head->prev = NULL;
  mf->tail->prev = mf->head;
  mf->tail->next = NULL;


  return mf;
}


MBFifo *
mf_cache_init (MBFifo * mf)
{
  int i = 0;



  mf->mbcache = kmalloc (sizeof (MABuff *) * (mf->mbc_size + 1), GFP_KERNEL);
  if (!mf->mbcache)
    {
      return NULL;
    }

  for (i = 0; i < mf->mbc_size; i++)
    {
      if (!mf_insert_mabuff (mf, i))
	{
	  mf_cache_cleanup (mf);
	  return NULL;
	}
    }


  return mf;
}


MBFifo *
mf_fifo_init (MBFifo * mf)
{
  mf->totalbytes = 0;
  mf->frags = 0;
  mf->count = 0;

  mf->maxfragsize = MA_HARD_MAXFRAGSIZE;
  mf->minfragsize = MA_HARD_MINFRAGSIZE;
  mf->maxfrags = MA_HARD_MAXFRAGS;

  mf->mbc_size = mf->maxfrags;
  mf->mbcache = NULL;

  mf->fragsize = mf->maxfragsize;

  mf->head = kmalloc (sizeof (MABuff), GFP_KERNEL);
  if (!mf->head)
    return NULL;
  memset (mf->head, 0, sizeof (MABuff));

  mf->tail = kmalloc (sizeof (MABuff), GFP_KERNEL);
  if (!mf->tail)
    {
      kfree (mf->head);
      return NULL;
    }
  memset (mf->tail, 0, sizeof (MABuff));

  mf->head->next = mf->tail;
  mf->head->prev = NULL;
  mf->tail->prev = mf->head;
  mf->tail->next = NULL;

  mf->rtok = 0;
  mf->wakeup_thr = mf->fragsize;

  init_waitqueue_head (&mf->read_wait);
  init_waitqueue_head (&mf->write_wait);

  return mf;
}


MBFifo *
mf_restart_cache (MBFifo * mf, unsigned new_mbc_size, unsigned new_fragsize)
{

  mf_cache_cleanup (mf);

  if (new_fragsize < mf->minfragsize)
    {
      new_fragsize = mf->minfragsize;
    }
  if (new_fragsize > mf->maxfragsize)
    {
      new_fragsize = mf->maxfragsize;
    }

  if (new_mbc_size > 128)
    {
      new_mbc_size = 128;
    }
  if (new_mbc_size < 4)
    {
      new_mbc_size = 4;
    }

  mf->mbc_size = new_mbc_size;
  mf->fragsize = new_fragsize;

  mf_cache_init (mf);

  mf->totalbytes = 0;
  mf->frags = 0;
  mf->count = 0;
  mf->ticks = 0;
  mf->rtok = 0;
  mf->wakeup_thr = mf->fragsize;


  return mf;
}


MBFifo *
mf_fifo_cleanup (MBFifo * mf)
{
  mf_cache_cleanup (mf);

  if (mf->tail)
    {
      kfree (mf->tail);
    }
  else
    {
    }

  if (mf->head)
    {
      kfree (mf->head);
    }
  else
    {
    }

  return mf;
}


int
mf_have_some_data (MBFifo * mf)
{
  if (mf->totalbytes > 0 && mf->frags > 0 && mf->rtok > mf->wakeup_thr
      && mf->rtok > 0)
    return 1;

  return 0;
}


int
mf_have_some_space (MBFifo * mf)
{
  if (mf->frags < mf->mbc_size)
    return 1;

  return 0;
}


MABuff *
mf_get_tail (MBFifo * mf)
{
  if (mf->head->next == mf->tail)
    return NULL;

  return mf->tail->prev;
}


MABuff *
mf_get_head (MBFifo * mf)
{
  if (mf->head->next == mf->tail)
    return NULL;

  return mf->head->next;
}


MABuff *
mf_append_tail (MBFifo * mf, MABuff * mb)
{
  MABuff *tmp = NULL;


  tmp = mf->tail->prev;

  tmp->next = mb;
  mb->prev = tmp;

  mf->tail->prev = mb;
  mb->next = mf->tail;

  mf->frags++;
  mb->bytes = 0;
  if (mf->frags > mf->mbc_size)
    {
    }

  return tmp;
}


MABuff *
mf_dequeue_head (MBFifo * mf)
{
  int pos = 0;
  MABuff *tmp = NULL;



  if (!mf->head)
    {
      return NULL;
    }

  tmp = mf->head->next;
  if (tmp == mf->tail)
    {
      return NULL;
    }

  if (tmp->bytes)

    tmp->bytes = 0;
  tmp->coff = 0;
  mf->frags--;

  mf->head->next = tmp->next;
  tmp->next->prev = mf->head;

  pos = tmp->mbcpos;

  mf->mbcache[pos] = mf->mbcache[0];
  mf->mbcache[pos]->mbcpos = pos;

  mf->mbcache[0] = tmp;
  tmp->mbcpos = 0;

  return mf->head->next;
}


MABuff *
mf_find_fragment (MBFifo * mf, size_t count)
{
  int i = 0, j = -1;



  if (mf->frags >= mf->mbc_size)
    return NULL;

  for (i = 0; i < mf->mbc_size; i++)
    {

      if (mf->mbcache[i])
	{
	  if ((mf->mbcache[i]->bytes) == 0
	      && (mf->mbcache[i]->buffersize) >= count)
	    {
	      return mf->mbcache[i];	// found allocated & empty mabuff
	    }
	}
      else if (j < 0)
	{
	  j = i;
	}
    }

  if (j < 0)
    {
      return NULL;
    }
  mf->mbcache[j] = mf_insert_mabuff (mf, j);

  return mf->mbcache[j];
}


size_t
ma_buffer_append_tail (MBFifo * mf, const char *buffer, size_t count,
		       int discard)
{
  int ret = 0;
  MABuff *mb = NULL;


  mb = mf->tail->prev;

  if (count > mf->maxfragsize * 16)
    {
      return -EFAULT;
    }

  if (!discard)
    {
      ret = copy_from_user (mb->buf.cbf + mb->bytes, buffer, count);
    }
  else
    {
      memset (mb->buf.cbf + mb->bytes, 0, count);
    }

  mb->bytes += count;
  mf->totalbytes += count;
  mf->count += count;

  return ret;
}


size_t
ma_buffer_append (MBFifo * mf, const char *buffer, size_t count, int discard)
{
  MABuff *mb = NULL;
  int ret = 0;


  down (&mf->mbcsem);


  mb = mf_get_tail (mf);

  if (mb && mb->buffersize - mb->bytes > count)
    {
      ret = ma_buffer_append_tail (mf, buffer, count, discard);
      if (!ret)
	{
	  ret = count;
	  goto out;
	}
      else
	{
	  ret = -EFAULT;
	  goto out;
	}
    }

  mb = mf_find_fragment (mf, count);

  if (mb)
    {

      count = count > mb->buffersize ? mb->buffersize : count;

      mf_append_tail (mf, mb);

      ret = ma_buffer_append_tail (mf, buffer, count, discard);
      if (ret)
	{
	  ret = -EFAULT;
	  goto out;
	}
      ret = count;
      goto out;
    }
  else
    {
      ret = 0;
      goto out;
    }


out:
  up (&mf->mbcsem);

  return ret;
}


size_t
ma_buffer_consume (MBFifo * mf, char *buffer, size_t count, int discard)
{
  MABuff *mb = NULL;
  int consumed = 0, todo = 0, cnt = 0;
  int ret = 0;


  down (&mf->mbcsem);

  if (mf->rtok <= 0)
    {
      goto out;
    }

  if (count > mf->rtok)
    todo = mf->rtok;
  else
    todo = count;

  mb = mf_get_head (mf);

  if (mb)
    {
    }
  else
    {
      goto out;
    }

  while (todo > 0)
    {

      if (!mb || mb == mf->tail)
	break;

      cnt = mb->bytes - mb->coff;

      if (cnt < 0)
	{
	  consumed = 0;
	  goto out;
	}
      if (mb->coff > mb->buffersize)
	{
	  consumed = 0;
	  goto out;
	}
      if (cnt > mb->buffersize)
	{
	  consumed = 0;
	  goto out;
	}

      if (cnt > todo)
	{
	  ret = 0;
	  if (!discard)
	    {
	      ret = copy_to_user (buffer, (mb->buf).cbf + mb->coff, todo);
	    }
	  if (ret)
	    {
	      consumed = -EFAULT;
	      goto out;
	    }

	  mf->totalbytes -= todo;
	  mf->rtok -= todo;
	  consumed += todo;
	  mb->coff += todo;	// consumer offset for this buffer
	  todo = 0;
	  goto out;
	}

      else
	{
	  ret = 0;
	  if (!discard)
	    {
	      ret = copy_to_user (buffer, (mb->buf).cbf + mb->coff, cnt);
	    }
	  if (ret)
	    {
	      consumed = -EFAULT;
	      goto out;
	    }

	  mb->bytes = 0;
	  mb->coff = 0;

	  mf_dequeue_head (mf);

	  buffer += cnt;
	  todo -= cnt;
	  consumed += cnt;
	  mf->totalbytes -= cnt;
	  mf->rtok -= cnt;
	}

      mb = mb->next;
    }

out:
  up (&mf->mbcsem);

  return consumed;
}


static ssize_t
maudio_read (struct file *file, char *buffer, size_t count, loff_t * ppos)
{
  ssize_t ret = -EINVAL;
  DECLARE_WAITQUEUE (wait, current);
  MState *s = NULL;


  s = (MState *) file->private_data;
  if (!s)
    {
      return -ENXIO;
    }


  if (ppos != &file->f_pos)
    {
      ret = -ESPIPE;
      goto out;
    }

  if (!buffer || !access_ok (VERIFY_WRITE, buffer, count))
    {
      ret = -EFAULT;
      goto out;
    }

  ret = 0;

  while (1)
    {

      ret = ma_buffer_consume (s->rmf, buffer, count, 0);

      if (ret == -EFAULT)
	goto out;

      if (ret == 0)
	{

	  if (file->f_flags & O_NONBLOCK)
	    {
	      ret = -EAGAIN;
	      goto out;
	    }
	  else
	    {
	      __set_current_state (TASK_INTERRUPTIBLE);
	      add_wait_queue (&s->rmf->read_wait, &wait);
	      schedule ();
	      remove_wait_queue (&s->rmf->read_wait, &wait);
	      set_current_state (TASK_RUNNING);
	      if (signal_pending (current))
		{
		  ret = -ERESTARTSYS;
		  goto out;
		}
	      continue;
	    }
	}

      break;
    }

out:
  return ret;
}


static ssize_t
maudio_write (struct file *file, const char *buffer, size_t count,
	      loff_t * ppos)
{
  ssize_t ret = 0, spd = 0;
  MState *s = NULL;
  DECLARE_WAITQUEUE (wait, current);


  s = (MState *) file->private_data;

  if (ppos != &file->f_pos)
    return -ESPIPE;

  if (!buffer || !access_ok (VERIFY_READ, buffer, count))
    return -EFAULT;


  while (1)
    {

      down (&s->wmf->devsem);
      if (s->wmf->read_dev)
	{
	  spd = 0;
	}
      else
	{
	  spd = 1;
	}
      up (&s->wmf->devsem);

      ret = ma_buffer_append (s->wmf, buffer, count, spd);

      if (ret == -EFAULT)
	return ret;

      if (ret == 0)
	{

	  if (file->f_flags & O_NONBLOCK)
	    {
	      ret = -EAGAIN;
	      goto out;
	    }
	  else
	    {
	      __set_current_state (TASK_INTERRUPTIBLE);
	      add_wait_queue (&s->wmf->write_wait, &wait);
	      schedule ();
	      remove_wait_queue (&s->wmf->write_wait, &wait);
	      set_current_state (TASK_RUNNING);
	      if (signal_pending (current))
		{
		  ret = -ERESTARTSYS;
		  goto out;
		}
	      continue;
	    }
	}

      break;
    }

out:
  return ret;
}


static unsigned int
maudio_poll (struct file *file, struct poll_table_struct *wait)
{
  unsigned int mask = 0;
  MState *s = NULL;


  s = (MState *) file->private_data;


  if (s->rmf)
    {
      poll_wait (file, &s->rmf->read_wait, wait);
      if (mf_have_some_data (s->rmf))
	mask |= POLLIN | POLLRDNORM;
    }

  if (s->wmf)
    {
      poll_wait (file, &s->wmf->write_wait, wait);
      if (mf_have_some_space (s->wmf))
	mask |= POLLOUT | POLLWRNORM;
    }

  return mask;
}


static int
maudio_mmap (struct file *file, struct vm_area_struct *vma)
{


  return -ENODEV;
}


static int
maudio_ioctl (struct inode *inode, struct file *file, unsigned int cmd,
	      unsigned long arg)
{
  unsigned long size;
  int val = 0, rate = 0;
  audio_buf_info abinfo;
  MState *s = NULL;


  s = (MState *) file->private_data;

  switch (cmd)
    {

    case OSS_GETVERSION:
      return put_user (SOUND_VERSION, (int *) arg);

    case SNDCTL_DSP_SYNC:
      return 0;

    case SNDCTL_DSP_SETDUPLEX:
      return 0;

    case SNDCTL_DSP_GETCAPS:
      return put_user (DSP_CAP_DUPLEX | DSP_CAP_REALTIME, (int *) arg);


    case SNDCTL_DSP_RESET:
      return 0;

    case SNDCTL_DSP_SPEED:

      if (get_user (val, (int *) arg))
	return -EFAULT;

      if (val >= 0)
	{

	  rate = val / HZ;
	  rate *= HZ;

	  if (rate < MA_HARD_MINRATE)
	    rate = MA_HARD_MINRATE;
	  else if (rate > MA_HARD_MAXRATE)
	    rate = MA_HARD_MAXRATE;

	  if (rate != val)
	    {
	    }

	  s->samplerate = rate;
	}
      else
	val = -EINVAL;

      s->bytespersec = rate * s->channels * s->depth;

      return put_user (val, (int *) arg);

    case SNDCTL_DSP_STEREO:

      if (get_user (val, (int *) arg))
	return -EFAULT;
      s->channels = val ? 2 : 1;
      s->bytespersec = s->samplerate * s->channels * s->depth;
      return 0;

    case SNDCTL_DSP_CHANNELS:

      if (get_user (val, (int *) arg))
	return -EFAULT;
      if (val != 0)
	{
	  s->channels = (val >= 2) ? 2 : 1;
	}
      s->bytespersec = s->samplerate * s->channels * s->depth;
      return put_user (s->channels, (int *) arg);

    case SNDCTL_DSP_GETFMTS:	// Returns a mask
      return put_user ((int) (AFMT_S16_LE | AFMT_U16_LE | AFMT_S8 | AFMT_U8),
		       (int *) arg);

    case SNDCTL_DSP_SETFMT:

      if (get_user (val, (int *) arg))
	return -EFAULT;


      if (val != AFMT_QUERY)
	{

	  if (val != AFMT_S16_LE && val != AFMT_U16_LE &&
	      val != AFMT_S8 && val != AFMT_U8)
	    val = AFMT_U8;
	  s->fmt = val;

	  if (val == AFMT_S16_LE || val == AFMT_U16_LE)
	    s->depth = 2;
	  else
	    s->depth = 1;

	}
      s->bytespersec = s->samplerate * s->channels * s->depth;
      return put_user (s->fmt, (int *) arg);


    case SNDCTL_DSP_POST:
      return 0;


    case SNDCTL_DSP_GETTRIGGER:
      printk (MODMSG "SNDCTL_DSP_GETTRIGGER\n");
      return 0;

    case SNDCTL_DSP_SETTRIGGER:
      printk (MODMSG "SNDCTL_DSP_SETTRIGGER\n");
      return 0;

    case SNDCTL_DSP_GETOSPACE:
      abinfo.fragsize = s->wmf->fragsize;
      abinfo.fragstotal = s->wmf->mbc_size;
      abinfo.fragments = s->wmf->mbc_size - s->wmf->frags;
      abinfo.bytes = s->wmf->frags * s->wmf->fragsize;
      return copy_to_user ((void *) arg, &abinfo,
			   sizeof (abinfo)) ? -EFAULT : 0;

    case SNDCTL_DSP_GETISPACE:
      abinfo.fragsize = s->rmf->fragsize;
      abinfo.fragstotal = s->rmf->mbc_size;
      abinfo.fragments = s->rmf->frags;
      abinfo.bytes = s->rmf->frags * s->rmf->fragsize;
      return copy_to_user ((void *) arg, &abinfo,
			   sizeof (abinfo)) ? -EFAULT : 0;

    case SNDCTL_DSP_NONBLOCK:
      file->f_flags |= O_NONBLOCK;
      return 0;

    case SNDCTL_DSP_GETODELAY:
      return 0;

    case SNDCTL_DSP_GETIPTR:
      return 0;

    case SNDCTL_DSP_GETOPTR:
      return 0;

    case SNDCTL_DSP_GETBLKSIZE:
      if (s->rmf)
	size = s->rmf->fragsize;
      else if (s->wmf)
	size = s->wmf->fragsize;
      else
	size = 0;
      return put_user (size, (int *) arg);

    case SNDCTL_DSP_SETFRAGMENT:
      return 0;

    case SNDCTL_DSP_SUBDIVIDE:
      return 0;

    case SOUND_PCM_READ_RATE:
      return put_user (s->samplerate, (int *) arg);

    case SOUND_PCM_READ_CHANNELS:
      return put_user (s->channels, (int *) arg);

    case SOUND_PCM_READ_BITS:
      return put_user ((s->fmt & (AFMT_S8 | AFMT_U8)) ? 8 : 16, (int *) arg);

    case SOUND_PCM_WRITE_FILTER:
    case SNDCTL_DSP_SETSYNCRO:
    case SOUND_PCM_READ_FILTER:
      return 0;

    default:
      return -EINVAL;
    }

  return 0;
}


static int
maudio_release (struct inode *inode, struct file *file)
{
  int minor = 0;
  MState *s = NULL;


  minor = MINOR (inode->i_rdev);

  lock_kernel ();

  s = (MState *) file->private_data;
  file->private_data = NULL;
  s->open_mode &= ~(FMODE_READ | FMODE_WRITE);

  if (minor == dsp_state.dev_audio)
    {
      printk (MODMSG "closing DSP device\n");
      if (file->f_mode & FMODE_READ)
	{
	  down (&global_readf.devsem);
	  global_readf.read_dev = NULL;
	  up (&global_readf.devsem);
	  s->rmf = NULL;
	}
      if (file->f_mode & FMODE_WRITE)
	{
	  down (&global_writef.devsem);
	  global_writef.write_dev = NULL;
	  up (&global_writef.devsem);
	  s->wmf = NULL;
	}
    }

  else if (minor == psd_state.dev_audio)
    {
      printk (MODMSG "closing PSD device\n");
      if (file->f_mode & FMODE_READ)
	{
	  down (&global_writef.devsem);
	  global_writef.read_dev = NULL;
	  up (&global_writef.devsem);
	  s->rmf = NULL;
	}
      if (file->f_mode & FMODE_WRITE)
	{
	  down (&global_readf.devsem);
	  global_readf.write_dev = NULL;
	  up (&global_readf.devsem);
	  s->wmf = NULL;
	}
    }

  else
    {
      printk (MODMSG "Oooops, I shouldn't have been opened at all!!!!!!\n");
      goto out_wakeup;
    }

out_wakeup:
  s->opens--;

  if (minor == dsp_state.dev_audio)
    {
      printk (MODMSG "CLOSED DSP device\n");
    }
  else
    {
      printk (MODMSG "CLOSED PSD device\n");
    }

  wake_up (s->open_wait);

  unlock_kernel ();

  return 0;
}


static int
maudio_open (struct inode *inode, struct file *file)
{
  int minor = 0, ret = 0;
  DECLARE_WAITQUEUE (wait, current);
  MState *s = NULL;


  if (!file || !inode)
    {
      ret = -EINVAL;
      goto out;
    }

  minor = MINOR (inode->i_rdev);


  if (minor == dsp_state.dev_audio)
    {
      printk (MODMSG "opening DSP device\n");
      s = &dsp_state;
    }
  else if (minor == psd_state.dev_audio)
    {
      printk (MODMSG "opening PSD device\n");
      s = &psd_state;
    }
  else
    {
      printk (MODMSG "nerd, what are you opening me for?\n");
      ret = -ENODEV;
      goto out;
    }

  if (down_interruptible (&s->opensem))
    return -ERESTARTSYS;


  while (s->opens)
    {
      if (file->f_flags & O_NONBLOCK)
	{
	  ret = -EBUSY;
	  goto out_up;
	}

      add_wait_queue (s->open_wait, &wait);
      __set_current_state (TASK_INTERRUPTIBLE);
      schedule ();
      remove_wait_queue (s->open_wait, &wait);
      set_current_state (TASK_RUNNING);
      if (signal_pending (current))
	{
	  ret = -ERESTARTSYS;
	  goto out_up;
	}
    }

  lock_kernel ();
  s->opens++;


  if (minor == dsp_state.dev_audio)
    {
      ms_init_device (s);

      if (file->f_mode & FMODE_READ)
	{
	  mf_restart_cache (&global_readf, global_readf.mbc_size,
			    global_readf.fragsize);
	  s->rmf = &global_readf;
	  down (&global_readf.devsem);
	  global_readf.read_dev = s;
	  up (&global_readf.devsem);
	}
      if (file->f_mode & FMODE_WRITE)
	{
	  mf_restart_cache (&global_writef, global_writef.mbc_size,
			    global_writef.fragsize);
	  s->wmf = &global_writef;
	  down (&global_readf.devsem);
	  global_writef.write_dev = s;
	  up (&global_readf.devsem);
	}
    }

  else if (minor == psd_state.dev_audio)
    {
      ms_init_device (s);

      if (file->f_mode & FMODE_READ)
	{
	  mf_restart_cache (&global_writef, global_writef.mbc_size,
			    global_writef.fragsize);
	  s->rmf = &global_writef;
	  down (&global_writef.devsem);
	  global_writef.read_dev = s;
	  up (&global_writef.devsem);
	}
      if (file->f_mode & FMODE_WRITE)
	{
	  mf_restart_cache (&global_readf, global_readf.mbc_size,
			    global_readf.fragsize);
	  s->wmf = &global_readf;
	  down (&global_writef.devsem);
	  global_readf.write_dev = s;
	  up (&global_writef.devsem);
	}

      s->samplerate = dsp_state.samplerate;
      s->channels = dsp_state.channels;
      s->depth = dsp_state.depth;
      s->bytespersec = s->channels * s->depth * s->samplerate;
      printk (MODMSG "cloned device settings: R %d  C %d  D %d  F %d\n",
	      s->samplerate, s->channels, s->depth, s->bytespersec);
    }

  s->open_mode |= file->f_mode & (FMODE_READ | FMODE_WRITE);
  file->private_data = s;

  unlock_kernel ();

  if (minor == dsp_state.dev_audio)
    {
      printk (MODMSG "opened DSP device\n");
    }
  else
    {
      printk (MODMSG "opened PSD device\n");
    }

out_up:
  up (&s->opensem);

out:
  return ret;
}


static struct file_operations maudio_audio_fops = {
  owner:THIS_MODULE,
  llseek:no_llseek,
  read:maudio_read,
  write:maudio_write,
  poll:maudio_poll,
  ioctl:maudio_ioctl,
  mmap:maudio_mmap,
  open:maudio_open,
  release:maudio_release,
};



void
trigger_fifo (MBFifo * mf)
{
  int count = 0, sinksrc = 0, spcdata = 0;
  volatile MState *ms = NULL;


  if (!down_trylock (&mf->devsem))
    {
      if (mf->read_dev)
	{
	  ms = mf->read_dev;
	  if (!mf->write_dev)
	    sinksrc++;
	}
      else
	{
	  ms = mf->write_dev;
	  sinksrc++;
	}
      up (&mf->devsem);
    }

  if (!ms)
    {
      return;
    }

  mf->ticks++;

  count = ms->bytespersec / HZ;
  if (mf->rtok < mf->frags * mf->fragsize)
    mf->rtok += count;

  if (mf_have_some_data (mf))
    {
      wake_up (&mf->read_wait);
      spcdata++;
    }

  if (mf_have_some_space (mf))
    {
      wake_up (&mf->write_wait);
      spcdata++;
    }

  if (sinksrc && spcdata)
    {
      wake_up (&maudiod_wait);
    }
}


void
mf_print_stat (MBFifo * mf, const char *name)
{
  size_t bs, ss, dp, cn, ticks = 0;

  bs = mf->count;

  ticks = mf->ticks;
  if (!ticks)
    ticks = 1;

  bs /= (ticks);
  bs *= HZ;

  if (mf->write_dev)
    {
      dp = (mf->write_dev)->depth;
      cn = (mf->write_dev)->channels;
    }
  else if (mf->read_dev)
    {
      dp = (mf->read_dev)->depth;
      cn = (mf->read_dev)->channels;
    }
  else
    {
      dp = 1;
      cn = 1;
      bs = 0;
    }

  ss = bs / dp / cn;

  printk (MODMSG
	  "%s STAT:  B %.6d  F %.3d  RATE: %.6u %.6u [b/s s/s]  time: %d [s]  total: %u [bytes]\n",
	  name, mf->totalbytes, mf->frags, bs, ss, maticks / HZ, mf->count);
}


#ifdef SLOWLIFE
static int mtdiv = 0;
#endif

void
ma_timer_func (unsigned long v)
{
  if (!(matimerflg & MA_TIMER_STOP))
    {
      matim.expires = jiffies + 1;
      add_timer (&matim);
    }

#ifdef SLOWLIFE
  mtdiv++;
  if (mtdiv % 300)
    return;
#endif

  maticks++;

  trigger_fifo (&global_readf);
  trigger_fifo (&global_writef);

  if (printstat && !(maticks % MA_TIMER_STATICS))
    {
      mf_print_stat (&global_readf, "READF");
      mf_print_stat (&global_writef, "WRTEF");

    }
}


int
handle_fifo_md (MBFifo * mf)
{
  int mask = 0, ret = 0;


  down (&mf->devsem);

  if (!mf->read_dev && mf->write_dev)
    {
      ma_buffer_consume (mf, NULL, mf->rtok, 1);
      mask |= 1;
    }
  else if (mf->read_dev && !mf->write_dev)
    {
      do
	{
	  ret = ma_buffer_append (mf, NULL, mf->fragsize, 1);
	}
      while (ret && ret != -EFAULT);
      mask |= 2;
    }
  else
    {

    }

  up (&mf->devsem);

  return mask;
}


int
mathread (void *arg)
{
  int rmask, wmask;


  printk (MODMSG "maudio thread starting\n");
  lock_kernel ();
  daemonize ();
  spin_lock_irq (&current->sigmask_lock);
  sigfillset (&current->blocked);
  recalc_sigpending (current);
  spin_unlock_irq (&current->sigmask_lock);
  unlock_kernel ();

  while (!(matimerflg & MA_TIMER_KILL_THREAD))
    {

      rmask = handle_fifo_md (&global_readf);
      wmask = handle_fifo_md (&global_writef);

      sprintf (current->comm, "maudiod %x %x", rmask, wmask);

      set_task_state (current, TASK_INTERRUPTIBLE);
      interruptible_sleep_on_timeout (&maudiod_wait, HZ);

      if (signal_pending (current))
	{
	  spin_lock_irq (&current->sigmask_lock);
	  flush_signals (current);
	  recalc_sigpending (current);
	  spin_unlock_irq (&current->sigmask_lock);
	}

      set_task_state (current, TASK_RUNNING);
    }

  printk (MODMSG "maudio thread exiting\n");

  return 0;
}


int __init
init_maudio (void)
{
  int ret = 0, devnr;


  printk (MODMSG "virtual sound card driver\n");

  init_waitqueue_head (&dsp_open_wait);
  init_waitqueue_head (&psd_open_wait);

  ms_init_device (&dsp_state);
  ms_init_device (&psd_state);
  init_MUTEX (&dsp_state.opensem);
  init_MUTEX (&psd_state.opensem);
  dsp_state.open_wait = &dsp_open_wait;
  psd_state.open_wait = &psd_open_wait;

  mf_fifo_init (&global_readf);
  mf_fifo_init (&global_writef);
  init_MUTEX (&global_readf.mbcsem);
  init_MUTEX (&global_writef.mbcsem);
  init_MUTEX (&global_readf.devsem);
  init_MUTEX (&global_writef.devsem);

  if ((devnr = register_sound_dsp (&maudio_audio_fops, -1)) < 0)
    {
      ret = devnr;
      goto err_dev1;
    }
  dsp_state.dev_audio = devnr;
  printk (MODMSG "registered sound device at node %d\n", devnr);

  if ((devnr = register_sound_dsp (&maudio_audio_fops, -1)) < 0)
    {
      ret = devnr;
      goto err_dev2;
    }
  psd_state.dev_audio = devnr;
  printk (MODMSG "registered sound device at node %d\n", devnr);

  init_timer (&matim);
  matim.function = &ma_timer_func;
  matim.expires = jiffies + 1;
  matim.data = 0;
  maticks = 1;
  matimerflg = 0;
  add_timer (&matim);

  kthrpid = kernel_thread (&mathread, NULL, 0);

  printk (MODMSG "initialisation success\n");
  ret = 0;
  goto out;

err_dev2:
  unregister_sound_dsp (dsp_state.dev_audio);

err_dev1:
  mf_fifo_cleanup (&global_readf);
  mf_fifo_cleanup (&global_writef);

  ms_cleanup_device (&dsp_state);
  ms_cleanup_device (&psd_state);

  printk (MODERR "initialisation error %d\n", ret);

out:
  return ret;
}

DECLARE_WAIT_QUEUE_HEAD (thrqueue);

void __exit
cleanup_maudio (void)
{
  lock_kernel ();

  printk (MODMSG "elapsed timer ticks %d\n", maticks);

  matimerflg |= MA_TIMER_STOP | MA_TIMER_KILL_THREAD;
  del_timer_sync (&matim);

  wake_up (&global_readf.read_wait);
  wake_up (&global_writef.read_wait);
  wake_up (&global_readf.write_wait);
  wake_up (&global_writef.write_wait);
  wake_up (&maudiod_wait);

  unregister_sound_dsp (dsp_state.dev_audio);
  printk (MODMSG "unregistered device %d\n", dsp_state.dev_audio);
  unregister_sound_dsp (psd_state.dev_audio);
  printk (MODMSG "unregistered device %d\n", psd_state.dev_audio);

  mf_fifo_cleanup (&global_readf);
  mf_fifo_cleanup (&global_writef);

  ms_cleanup_device (&dsp_state);
  ms_cleanup_device (&psd_state);

  printk (MODMSG "exiting maudiod\n");
  while (kthrpid && find_task_by_pid (kthrpid))
    schedule ();

  printk (MODMSG "removed driver\n");

  unlock_kernel ();
}


MODULE_AUTHOR ("Paul Starzetz paul@starzetz.de");
MODULE_DESCRIPTION ("Mirroring audio driver version " MVERSION);
MODULE_LICENSE ("GPL");

module_init (init_maudio);
module_exit (cleanup_maudio);
